/*-
 * Copyright (c) 2008 Weongyo Jeong <weongyo@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    similar to the "NO WARRANTY" disclaimer below ("Disclaimer") and any
 *    redistribution must be conditioned upon including a substantially
 *    similar Disclaimer requirement for further binary redistribution.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF NONINFRINGEMENT, MERCHANTIBILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */

#ifndef _IF_BWN_SSB_H
#define	_IF_BWN_SSB_H

struct ssb_softc;
struct ssb_dev_softc;

#define	PCI_DEVICE_ID_BCM4401		0x4401
#define	PCI_DEVICE_ID_BCM4401B0		0x4402
#define	PCI_DEVICE_ID_BCM4401B1		0x170c
#define	SSB_PCIR_BAR		PCIR_BAR(0)

/*
 * ChipCommon registers.
 */
#define	SSB_CC_CHIPID			0x0000
#define	SSB_CC_ID(id)			(id & 0x0000ffff)
#define	SSB_CC_REV(id)			((id & 0x000f0000) >> 16)
#define	SSB_CC_PKG(id)			((id & 0x00f00000) >> 20)
#define	SSB_CC_NCORES(id)		((id & 0x0f000000) >> 24)
#define	SSB_CC_CAPS			0x0004
#define	SSB_CC_CAPS_PWCTL		0x00040000
#define	SSB_CC_CAPS_PMU			0x10000000	/* PMU (rev >= 20) */
#define	SSB_CC_CHIPCTL			0x0028		/* rev >= 11 */
#define	SSB_CC_CHIPSTAT			0x002C		/* rev >= 11 */
#define	SSB_CC_BCAST_ADDR		0x0050		/* Broadcast Address */
#define	SSB_CC_BCAST_DATA		0x0054		/* Broadcast Data */
#define	SSB_CC_PLLONDELAY		0x00B0		/* Rev >= 4 only */
#define	SSB_CC_FREFSELDELAY		0x00B4		/* Rev >= 4 only */
#define	SSB_CC_CLKSLOW			0x00b8		/* 6 <= Rev <= 9 only */
#define	SSB_CC_CLKSLOW_SRC		0x00000007
#define	SSB_CC_CLKSLOW_SRC_CRYSTAL	0x00000001
#define	SSB_CC_CLKSLOW_FSLOW		0x00000800
#define	SSB_CC_CLKSLOW_IPLL		0x00001000
#define	SSB_CC_CLKSLOW_ENXTAL		0x00002000
#define	SSB_CC_CLKSYSCTL		0x00C0		/* Rev >= 3 only */
#define	SSB_CC_CLKCTLSTATUS		0x01e0
#define	SSB_CC_CLKCTLSTATUS_HT		0x00010000
#define	SSB_CC_PMUCTL			0x0600		/* PMU control */
#define	SSB_CC_PMUCTL_ILP		0xffff0000	/* mask */
#define	SSB_CC_PMUCTL_NOILP		0x00000200
#define	SSB_CC_PMUCTL_XF		0x0000007c	/* crystal freq */
#define	SSB_CC_PMUCTL_XF_VAL(id)	((id & 0x0000007c) >> 2)
#define	SSB_CC_PMUCAPS			0x0604
#define	SSB_CC_PMUCAPS_REV		0x000000ff
#define	SSB_CC_PMU_MINRES		0x0618
#define	SSB_CC_PMU_MAXRES		0x061c
#define	SSB_CC_PMU_TABSEL		0x0620
#define	SSB_CC_PMU_DEPMSK		0x0624
#define	SSB_CC_PMU_UPDNTM		0x0628
#define	SSB_CC_PLLCTL_ADDR		0x0660
#define	SSB_CC_PLLCTL_DATA		0x0664

#define	SSB_CC_PMU0_PLL0		0
#define	SSB_CC_PMU0_PLL0_PDIV_MSK	0x00000001
#define	SSB_CC_PMU0_PLL0_PDIV_FREQ	25000
#define	SSB_CC_PMU0_PLL1		1
#define	SSB_CC_PMU0_PLL1_IMSK		0xf0000000
#define	SSB_CC_PMU0_PLL1_FMSK		0x0fffff00
#define	SSB_CC_PMU0_PLL1_STOPMOD	0x00000040
#define	SSB_CC_PMU0_PLL2		2
#define	SSB_CC_PMU0_PLL2_IMSKHI		0x0000000f
#define	SSB_CC_PMU1_PLL0		0
#define	SSB_CC_PMU1_PLL0_P1DIV		0x00f00000
#define	SSB_CC_PMU1_PLL0_P2DIV		0x0f000000
#define	SSB_CC_PMU1_PLL2		2
#define	SSB_CC_PMU1_PLL2_NDIVMODE	0x000e0000
#define	SSB_CC_PMU1_PLL2_NDIVINT	0x1ff00000
#define	SSB_CC_PMU1_PLL3		3
#define	SSB_CC_PMU1_PLL3_NDIVFRAC	0x00ffffff
#define	SSB_CC_PMU1_PLL5		5
#define	SSB_CC_PMU1_PLL5_CLKDRV		0xffffff00

#define	SSB_CC_PMU1_PLLTAB_ENTRY					\
{									\
	{ 12000,  1, 3, 22,  0x9, 0xffffef },				\
	{ 13000,  2, 1,  6,  0xb, 0x483483 },				\
	{ 14400,  3, 1, 10,  0xa, 0x1c71c7 },				\
	{ 15360,  4, 1,  5,  0xb, 0x755555 },				\
	{ 16200,  5, 1, 10,  0x5, 0x6e9e06 },				\
	{ 16800,  6, 1, 10,  0x5, 0x3cf3cf },				\
	{ 19200,  7, 1,  9,  0x5, 0x17b425 },				\
	{ 19800,  8, 1, 11,  0x4, 0xa57eb },				\
	{ 20000,  9, 1, 11,  0x4, 0 },					\
	{ 24000, 10, 3, 11,  0xa, 0 },					\
	{ 25000, 11, 5, 16,  0xb, 0 },					\
	{ 26000, 12, 1,  2, 0x10, 0xec4ec4 },				\
	{ 30000, 13, 3,  8,  0xb, 0 },					\
	{ 38400, 14, 1,  5,  0x4, 0x955555 },				\
	{ 40000, 15, 1,  2,  0xb, 0 }					\
}

#define	SSB_CC_PMU0_PLLTAB_ENTRY					\
{									\
	{ 12000,  1, 73, 349525, }, { 13000,  2, 67, 725937, },		\
	{ 14400,  3, 61, 116508, }, { 15360,  4, 57, 305834, },		\
	{ 16200,  5, 54, 336579, }, { 16800,  6, 52, 399457, },		\
	{ 19200,  7, 45, 873813, }, { 19800,  8, 44, 466033, },		\
	{ 20000,  9, 44, 0,      }, { 25000, 10, 70, 419430, },		\
	{ 26000, 11, 67, 725937, }, { 30000, 12, 58, 699050, },		\
	{ 38400, 13, 45, 873813, }, { 40000, 14, 45, 0,      },		\
}

#define	SSB_CC_PMU_4325_BURST		1
#define	SSB_CC_PMU_4325_CLBURST		3
#define	SSB_CC_PMU_4325_LN		10
#define	SSB_CC_PMU_4325_CRYSTAL		13
#define	SSB_CC_PMU_4325_RX_PWR		15
#define	SSB_CC_PMU_4325_TX_PWR		16
#define	SSB_CC_PMU_4325_LOGEN_PWR	18
#define	SSB_CC_PMU_4325_AFE_PWR		19
#define	SSB_CC_PMU_4325_BBPLL_PWR	20
#define	SSB_CC_PMU_4325_HT		21
#define	SSB_CC_PMU_4328_EXT_SWITCH_PWM	0
#define	SSB_CC_PMU_4328_BB_SWITCH_PWM	1
#define	SSB_CC_PMU_4328_BB_SWITCH_BURST	2
#define	SSB_CC_PMU_4328_BB_EXT_SWITCH_BURST	3
#define	SSB_CC_PMU_4328_ILP_REQUEST	4
#define	SSB_CC_PMU_4328_RADSWITCH_PWM	5	/* radio switch */
#define	SSB_CC_PMU_4328_RADSWITCH_BURST	6
#define	SSB_CC_PMU_4328_ROM_SWITCH	7
#define	SSB_CC_PMU_4328_PA_REF		8
#define	SSB_CC_PMU_4328_RADIO		9
#define	SSB_CC_PMU_4328_AFE		10
#define	SSB_CC_PMU_4328_PLL		11
#define	SSB_CC_PMU_4328_BG_FILTBYP	12
#define	SSB_CC_PMU_4328_TX_FILTBYP	13
#define	SSB_CC_PMU_4328_RX_FILTBYP	14
#define	SSB_CC_PMU_4328_CRYSTAL_PU	15
#define	SSB_CC_PMU_4328_CRYSTAL_EN	16
#define	SSB_CC_PMU_4328_BB_PLL_FILTBYP	17
#define	SSB_CC_PMU_4328_RF_PLL_FILTBYP	18
#define	SSB_CC_PMU_4328_BB_PLL_PU	19
#define	SSB_CC_PMU_5354_BB_PLL_PU	19

#define	SSB_CC_PMU_4325_RES_UPDOWN					\
{									\
	{ SSB_CC_PMU_4325_CRYSTAL, 0x1501 }				\
}

#define	SSB_CC_PMU_4325_RES_DEPEND					\
{									\
	{ SSB_CC_PMU_4325_HT, SSB_CC_PMU_DEP_ADD,			\
	  ((1 << SSB_CC_PMU_4325_RX_PWR) |				\
	   (1 << SSB_CC_PMU_4325_TX_PWR) |				\
	   (1 << SSB_CC_PMU_4325_LOGEN_PWR) |				\
	   (1 << SSB_CC_PMU_4325_AFE_PWR)) }				\
}

#define	SSB_CC_PMU_4328_RES_UPDOWN					\
{									\
	{ SSB_CC_PMU_4328_EXT_SWITCH_PWM, 0x0101 },			\
	{ SSB_CC_PMU_4328_BB_SWITCH_PWM, 0x1f01 },			\
	{ SSB_CC_PMU_4328_BB_SWITCH_BURST, 0x010f },			\
	{ SSB_CC_PMU_4328_BB_EXT_SWITCH_BURST, 0x0101 },		\
	{ SSB_CC_PMU_4328_ILP_REQUEST, 0x0202 },			\
	{ SSB_CC_PMU_4328_RADSWITCH_PWM, 0x0f01 },			\
	{ SSB_CC_PMU_4328_RADSWITCH_BURST, 0x0f01 },			\
	{ SSB_CC_PMU_4328_ROM_SWITCH, 0x0101 },				\
	{ SSB_CC_PMU_4328_PA_REF, 0x0f01 },				\
	{ SSB_CC_PMU_4328_RADIO, 0x0f01 },				\
	{ SSB_CC_PMU_4328_AFE, 0x0f01 },				\
	{ SSB_CC_PMU_4328_PLL, 0x0f01 },				\
	{ SSB_CC_PMU_4328_BG_FILTBYP, 0x0101 },				\
	{ SSB_CC_PMU_4328_TX_FILTBYP, 0x0101 },				\
	{ SSB_CC_PMU_4328_RX_FILTBYP, 0x0101 },				\
	{ SSB_CC_PMU_4328_CRYSTAL_PU, 0x0101 },				\
	{ SSB_CC_PMU_4328_CRYSTAL_EN, 0xa001 },				\
	{ SSB_CC_PMU_4328_BB_PLL_FILTBYP, 0x0101 },			\
	{ SSB_CC_PMU_4328_RF_PLL_FILTBYP, 0x0101 },			\
	{ SSB_CC_PMU_4328_BB_PLL_PU, 0x0701 },				\
}

#define	SSB_CC_PMU_4328_RES_DEPEND					\
{									\
	{ SSB_CC_PMU_4328_ILP_REQUEST, SSB_CC_PMU_DEP_SET,		\
	  ((1 << SSB_CC_PMU_4328_EXT_SWITCH_PWM) |			\
	   (1 << SSB_CC_PMU_4328_BB_SWITCH_PWM)) },			\
}

#define	SSB_CC_CHST_4325_PMUTOP_2B	0x00000200

/* device ID */
#define	SSB_DEV_CHIPCOMMON		0x800
#define	SSB_DEV_ILINE20			0x801
#define	SSB_DEV_SDRAM			0x803
#define	SSB_DEV_PCI			0x804
#define	SSB_DEV_MIPS			0x805
#define	SSB_DEV_ETHERNET		0x806
#define	SSB_DEV_V90			0x807
#define	SSB_DEV_USB11_HOSTDEV		0x808
#define	SSB_DEV_ADSL			0x809
#define	SSB_DEV_ILINE100		0x80a
#define	SSB_DEV_IPSEC			0x80b
#define	SSB_DEV_PCMCIA			0x80d
#define	SSB_DEV_INTERNAL_MEM		0x80e
#define	SSB_DEV_MEMC_SDRAM		0x80f
#define	SSB_DEV_EXTIF			0x811
#define	SSB_DEV_80211			0x812
#define	SSB_DEV_MIPS_3302		0x816
#define	SSB_DEV_USB11_HOST		0x817
#define	SSB_DEV_USB11_DEV		0x818
#define	SSB_DEV_USB20_HOST		0x819
#define	SSB_DEV_USB20_DEV		0x81a
#define	SSB_DEV_SDIO_HOST		0x81b
#define	SSB_DEV_ROBOSWITCH		0x81c
#define	SSB_DEV_PARA_ATA		0x81d
#define	SSB_DEV_SATA_XORDMA		0x81e
#define	SSB_DEV_ETHERNET_GBIT		0x81f
#define	SSB_DEV_PCIE			0x820
#define	SSB_DEV_MIMO_PHY		0x821
#define	SSB_DEV_SRAM_CTRLR		0x822
#define	SSB_DEV_MINI_MACPHY		0x823
#define	SSB_DEV_ARM_1176		0x824
#define	SSB_DEV_ARM_7TDMI		0x825

#define	SSB_VENDOR_BROADCOM		0x4243

#define	SSB_BAR0			0x80
#define	SSB_IRQMASK			0x94
#define	SSB_GPIO_IN			0xb0
#define	SSB_GPIO_OUT			0xb4
#define	SSB_GPIO_OUT_EN			0xb8
#define	SSB_GPIO_CRYSTAL		0x40
#define	SSB_GPIO_PLL			0x80

#define	SSB_REGWIN(x)			(SSB_ENUM_START + ((x) * SSB_CORESIZE))
#define	SSB_CORESIZE			0x1000
#define	SSB_MAX_CORES					\
	((SSB_ENUM_END - SSB_ENUM_START) / SSB_CORESIZE)
#define	SSB_ENUM_START			0x18000000U
#define	SSB_ENUM_END			0x18010000U

#define	SSB_PCI_DMA			0x40000000U
#define	SSB_TPS				0x0f18
#define	SSB_TPS_BPFLAG			0x0000003f
#define	SSB_IAS				0x0f90     /* Initiator Agent State */
#define	SSB_IAS_INBAND_ERR		0x00020000
#define	SSB_IAS_TIMEOUT			0x00040000
#define	SSB_INTR_MASK			0x0f94
#define	SSB_TGSLOW			0x0f98
#define	SSB_TGSLOW_RESET		0x00000001	/* target state low */
#define	SSB_TGSLOW_REJECT_22		0x00000002
#define	SSB_TGSLOW_REJECT_23		0x00000004
#define	SSB_TGSLOW_CLOCK		0x00010000
#define	SSB_TGSLOW_FGC			0x00020000
#define	SSB_TGSHIGH			0x0f9c
#define	SSB_TGSHIGH_SERR		0x00000001
#define	SSB_TGSHIGH_BUSY		0x00000004
#define	SSB_TGSHIGH_DMA64		0x10000000
#define	SSB_IMCFGLO			0x0fa8
#define	SSB_IMCFGLO_SERTO		0x00000007
#define	SSB_IMCFGLO_REQTO		0x00000070
#define	SSB_IDLOW			0x0ff8
#define	SSB_IDLOW_SSBREV		0xf0000000
#define	SSB_IDLOW_SSBREV_22		0x00000000
#define	SSB_IDLOW_SSBREV_23		0x10000000
#define	SSB_IDLOW_SSBREV_24		0x40000000
#define	SSB_IDLOW_SSBREV_25		0x50000000
#define	SSB_IDLOW_SSBREV_26		0x60000000
#define	SSB_IDLOW_SSBREV_27		0x70000000
#define	SSB_IDHIGH			0x0ffc
#define	SSB_IDHIGH_REVLO		0x0000000f
#define	SSB_IDHIGH_REVHI		0x00007000
#define	SSB_IDHIGH_CORECODE(id)		((id & 0x8ff0) >> 4)
#define	SSB_IDHIGH_REV(id)					\
	((id & SSB_IDHIGH_REVLO) | ((id & SSB_IDHIGH_REVHI) >> 8))
#define	SSB_IDHIGH_VENDOR(id)		((id & 0xffff0000) >> 16)

#define	SSB_SPROMSIZE_R123		64
#define	SSB_SPROMSIZE_R4		220
#define	SSB_SPROM_BASE			0x1000
#define	SSB_SPROM_REV_CRC		0xff00
#define	SSB_SPROM1_MAC_80211BG		0x1048
#define	SSB_SPROM1_MAC_ETH		0x104e
#define	SSB_SPROM1_MAC_80211A		0x1054
#define	SSB_SPROM1_ETHPHY		0x105a
#define	SSB_SPROM1_ETHPHY_MII_ETH0	0x001f
#define	SSB_SPROM1_ETHPHY_MII_ETH1	0x03e0
#define	SSB_SPROM1_ETHPHY_MDIO_ETH0	(1 << 14)
#define	SSB_SPROM1_ETHPHY_MDIO_ETH1	(1 << 15)
#define	SSB_SPROM1_BOARDINFO		0x105c
#define	SSB_SPROM1_BOARDINFO_BREV	0x00ff
#define	SSB_SPROM1_BOARDINFO_CCODE	0x0f00
#define	SSB_SPROM1_BOARDINFO_ANTBG	0x3000
#define	SSB_SPROM1_BOARDINFO_ANTA	0xc000
#define	SSB_SPROM1_PA0B0		0x105e
#define	SSB_SPROM1_PA0B1		0x1060
#define	SSB_SPROM1_PA0B2		0x1062
#define	SSB_SPROM1_GPIOA		0x1064
#define	SSB_SPROM1_GPIOA_P0		0x00ff
#define	SSB_SPROM1_GPIOA_P1		0xff00
#define	SSB_SPROM1_GPIOB		0x1066
#define	SSB_SPROM1_GPIOB_P2		0x00ff
#define	SSB_SPROM1_GPIOB_P3		0xff00
#define	SSB_SPROM1_MAXPWR		0x1068
#define	SSB_SPROM1_MAXPWR_BG		0x00ff
#define	SSB_SPROM1_MAXPWR_A		0xff00
#define	SSB_SPROM1_PA1B0		0x106a
#define	SSB_SPROM1_PA1B1		0x106c
#define	SSB_SPROM1_PA1B2		0x106e
#define	SSB_SPROM1_TSSI		0x1070
#define	SSB_SPROM1_TSSI_BG		0x00ff
#define	SSB_SPROM1_TSSI_A		0xff00
#define	SSB_SPROM1_BFLOW		0x1072
#define	SSB_SPROM1_AGAIN		0x1074
#define	SSB_SPROM1_AGAIN_BG		0x00ff
#define	SSB_SPROM1_AGAIN_A		0xff00
#define	SSB_SPROM2_BFHIGH		0x1038
#define	SSB_SPROM3_MAC_80211BG		0x104a
#define	SSB_SPROM4_MAC_80211BG		0x104c
#define	SSB_SPROM4_ETHPHY		0x105a
#define	SSB_SPROM4_ETHPHY_ET0A		0x001f
#define	SSB_SPROM4_ETHPHY_ET1A		0x03e0
#define	SSB_SPROM4_CCODE		0x1052
#define	SSB_SPROM4_ANTAVAIL		0x105d
#define	SSB_SPROM4_ANTAVAIL_A		0x00ff
#define	SSB_SPROM4_ANTAVAIL_BG		0xff00
#define	SSB_SPROM4_BFLOW		0x1044
#define	SSB_SPROM4_AGAIN01		0x105e
#define	SSB_SPROM4_AGAIN0		0x00ff
#define	SSB_SPROM4_AGAIN1		0xff00
#define	SSB_SPROM4_AGAIN23		0x1060
#define	SSB_SPROM4_AGAIN2		0x00ff
#define	SSB_SPROM4_AGAIN3		0xff00
#define	SSB_SPROM4_BFHIGH		0x1046
#define	SSB_SPROM4_MAXP_BG		0x1080
#define	SSB_SPROM4_MAXP_BG_MASK		0x00ff
#define	SSB_SPROM4_TSSI_BG		0xff00
#define	SSB_SPROM4_MAXP_A		0x108a
#define	SSB_SPROM4_MAXP_A_MASK		0x00ff
#define	SSB_SPROM4_TSSI_A		0xff00
#define	SSB_SPROM4_GPIOA		0x1056
#define	SSB_SPROM4_GPIOA_P0		0x00ff
#define	SSB_SPROM4_GPIOA_P1		0xff00
#define	SSB_SPROM4_GPIOB		0x1058
#define	SSB_SPROM4_GPIOB_P2		0x00ff
#define	SSB_SPROM4_GPIOB_P3		0xff00
#define	SSB_SPROM5_BFLOW		0x104a
#define	SSB_SPROM5_BFHIGH		0x104c
#define	SSB_SPROM5_MAC_80211BG		0x1052
#define	SSB_SPROM5_CCODE		0x1044
#define	SSB_SPROM5_GPIOA		0x1076
#define	SSB_SPROM5_GPIOA_P0		0x00ff
#define	SSB_SPROM5_GPIOA_P1		0xff00
#define	SSB_SPROM5_GPIOB		0x1078
#define	SSB_SPROM5_GPIOB_P2		0x00ff
#define	SSB_SPROM5_GPIOB_P3		0xff00
#define	SSB_SPROM8_BFLOW		0x1084
#define	SSB_SPROM8_BFHIGH		0x1086
#define	SSB_SPROM8_CCODE		0x1092
#define	SSB_SPROM8_ANTAVAIL		0x109c
#define	SSB_SPROM8_ANTAVAIL_A		0xff00
#define	SSB_SPROM8_ANTAVAIL_BG		0x00ff
#define	SSB_SPROM8_AGAIN01		0x109e
#define	SSB_SPROM8_AGAIN0		0x00ff
#define	SSB_SPROM8_AGAIN1		0xff00
#define	SSB_SPROM8_AGAIN23		0x10a0
#define	SSB_SPROM8_AGAIN2		0x00ff
#define	SSB_SPROM8_AGAIN3		0xff00
#define	SSB_SPROM8_GPIOA		0x1096
#define	SSB_SPROM8_GPIOA_P0		0x00ff
#define	SSB_SPROM8_GPIOA_P1		0xff00
#define	SSB_SPROM8_GPIOB		0x1098
#define	SSB_SPROM8_GPIOB_P2		0x00ff
#define	SSB_SPROM8_GPIOB_P3		0xff00
#define	SSB_SPROM8_MAXP_BG		0x10c0
#define	SSB_SPROM8_MAXP_BG_MASK		0x00ff
#define	SSB_SPROM8_TSSI_BG		0xff00
#define	SSB_SPROM8_MAXP_A		0x10c8
#define	SSB_SPROM8_MAXP_A_MASK		0x00ff
#define	SSB_SPROM8_TSSI_A		0xff00

#define	SSB_BOARDVENDOR_DELL		0x1028
#define	SSB_BOARDVENDOR_BCM		0x14e4
#define	SSB_BOARD_BCM4309G		0x0421
#define	SSB_BOARD_MP4318		0x044a
#define	SSB_BOARD_BU4306		0x0416
#define	SSB_BOARD_BU4309		0x040a

#define	SSB_PCICORE_BCAST_ADDR		SSB_CC_BCAST_ADDR
#define	SSB_PCICORE_BCAST_DATA		SSB_CC_BCAST_DATA
#define	SSB_PCICORE_SBTOPCI2		0x0108
#define	SSB_PCICORE_MDIO_CTL		0x0128
#define	SSB_PCICORE_MDIO_DATA		0x012c
#define	SSB_PCICORE_SBTOPCI_PREF	0x00000004
#define	SSB_PCICORE_SBTOPCI_BURST	0x00000008
#define	SSB_PCICORE_SBTOPCI_MRM		0x00000020

enum {
	SSB_SPROM1CCODE_WORLD,
	SSB_SPROM1CCODE_THAILAND,
	SSB_SPROM1CCODE_ISRAEL,
	SSB_SPROM1CCODE_JORDAN,
	SSB_SPROM1CCODE_CHINA,
	SSB_SPROM1CCODE_JAPAN,
	SSB_SPROM1CCODE_USA_CANADA_ANZ,
	SSB_SPROM1CCODE_EUROPE,
	SSB_SPROM1CCODE_USA_LOW,
	SSB_SPROM1CCODE_JAPAN_HIGH,
	SSB_SPROM1CCODE_ALL,
	SSB_SPROM1CCODE_NONE,
};

#define	SSB_READ_4(ssb, reg)		\
	bus_space_read_4((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh, (reg))
#define	SSB_READ_2(ssb, reg)		\
	bus_space_read_2((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh, (reg))
#define	SSB_READ_MULTI_1(ssb, reg, addr, count)				\
	bus_space_read_multi_1((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh,	\
	    (reg), (addr), (count))
#define	SSB_READ_MULTI_2(ssb, reg, addr, count)				\
	bus_space_read_multi_2((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh,	\
	    (reg), (addr), (count))
#define	SSB_READ_MULTI_4(ssb, reg, addr, count)				\
	bus_space_read_multi_4((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh,	\
	    (reg), (addr), (count))

#define	SSB_WRITE_4(ssb, reg, val)	\
	bus_space_write_4((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh, (reg), (val))
#define	SSB_WRITE_2(ssb, reg, val)	\
	bus_space_write_2((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh, (reg), (val))
#define	SSB_WRITE_MULTI_1(ssb, reg, addr, count)			\
	bus_space_write_multi_1((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh,	\
	    (reg), (addr), (count))
#define	SSB_WRITE_MULTI_2(ssb, reg, addr, count)			\
	bus_space_write_multi_2((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh,	\
	    (reg), (addr), (count))
#define	SSB_WRITE_MULTI_4(ssb, reg, addr, count)			\
	bus_space_write_multi_4((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh,	\
	    (reg), (addr), (count))

#define	SSB_BARRIER(ssb, flags)						\
	bus_space_barrier((ssb)->ssb_mem_bt, (ssb)->ssb_mem_bh, (0),	\
	    (0), (flags))

#define	SSB_SETBITS_4(ssb, reg, bits)	\
	SSB_WRITE_4((ssb), (reg), SSB_READ_4((ssb), (reg)) | (bits))
#define	SSB_SETBITS_2(ssb, reg, bits)	\
	SSB_WRITE_2((ssb), (reg), SSB_READ_2((ssb), (reg)) | (bits))

#define	SSB_FILT_SETBITS_4(ssb, reg, filt, bits) \
	SSB_WRITE_4((ssb), (reg), (SSB_READ_4((ssb), (reg)) & (filt)) | (bits))
#define	SSB_FILT_SETBITS_2(ssb, reg, filt, bits)	\
	SSB_WRITE_2((ssb), (reg), (SSB_READ_2((ssb), (reg)) & (filt)) | (bits))

#define	SSB_CLRBITS_4(ssb, reg, bits)	\
	SSB_WRITE_4((ssb), (reg), SSB_READ_4((ssb), (reg)) & ~(bits))
#define	SSB_CLRBITS_2(ssb, reg, bits)	\
	SSB_WRITE_2((ssb), (reg), SSB_READ_2((ssb), (reg)) & ~(bits))

#define	SSB_CC_READ32(scc, offset) \
	ssb_read_4((scc)->scc_dev, offset)
#define	SSB_CC_WRITE32(scc, offset, val) \
	ssb_write_4((scc)->scc_dev, offset, val)
#define	SSB_CC_MASK32(scc, offset, mask) \
	SSB_CC_WRITE32(scc, offset, SSB_CC_READ32(scc, offset) & (mask))
#define	SSB_CC_SET32(scc, offset, set) \
	SSB_CC_WRITE32(scc, offset, SSB_CC_READ32(scc, offset) | (set))
#define	SSB_CC_MASKSET32(scc, offset, mask, set)	\
	SSB_CC_WRITE32(scc, offset,			\
	    (SSB_CC_READ32(scc, offset) & (mask)) | (set))

enum ssb_type {
	SSB_TYPE_SSB,
	SSB_TYPE_PCI,
	SSB_TYPE_PCMCIA,
};

enum ssb_clock {
	SSB_CLOCK_DYNAMIC,
	SSB_CLOCK_SLOW,
	SSB_CLOCK_FAST,
};

enum ssb_clksrc {
	SSB_CC_CLKSRC_PCI,
	SSB_CC_CLKSRC_CRYSTAL,
	SSB_CC_CLKSRC_LOWPW,
};

#define	SSB_CC_PMU0_DEFAULT_XTALFREQ	20000
#define	SSB_CC_PMU1_DEFAULT_FREQ	15360

struct ssb_cc_pmu0_plltab {
	uint16_t		freq;	/* in kHz.*/
	uint8_t			xf;	/* crystal frequency */
	uint8_t			wb_int;
	uint32_t		wb_frac;
};

struct ssb_cc_pmu1_plltab {
	uint16_t		freq;
	uint8_t			xf;
	uint8_t			p1div;
	uint8_t			p2div;
	uint8_t			ndiv_int;
	uint32_t		ndiv_frac;
};

struct ssb_cc_pmu_res_updown {
	uint8_t			res;
	uint16_t		updown;
};

#define	SSB_CC_PMU_DEP_SET	1
#define	SSB_CC_PMU_DEP_ADD	2
#define	SSB_CC_PMU_DEP_REMOVE	3

struct ssb_cc_pmu_res_depend {
	uint8_t			res;
	uint8_t			task;
	uint32_t		depend;
};

/* Information about the PCB the circuitry is soldered on. */
struct ssb_board {
	uint16_t		vendor;
	uint16_t		type;
	uint16_t		rev;
};

struct ssb_sprom {
	uint8_t			rev;		/* revision */
	uint8_t			mac_80211bg[6];	/* address for 802.11b/g */
	uint8_t			mac_eth[6];	/* address for Ethernet */
	uint8_t			mac_80211a[6];	/* address for 802.11a */
	uint8_t			mii_eth0;	/* MII address for eth0 */
	uint8_t			mii_eth1;	/* MII address for eth1 */
	uint8_t			mdio_eth0;	/* MDIO for eth0 */
	uint8_t			mdio_eth1;	/* MDIO for eth1 */
	uint8_t			brev;		/* board revision */
	uint8_t			ccode;		/* Country Code */
	uint8_t			ant_a;		/* A-PHY antenna */
	uint8_t			ant_bg;		/* B/G-PHY antenna */
	uint16_t		pa0b0;
	uint16_t		pa0b1;
	uint16_t		pa0b2;
	uint16_t		pa1b0;
	uint16_t		pa1b1;
	uint16_t		pa1b2;
	uint8_t			gpio0;
	uint8_t			gpio1;
	uint8_t			gpio2;
	uint8_t			gpio3;
	uint16_t		maxpwr_a;	/* A-PHY Max Power */
	uint16_t		maxpwr_bg;	/* BG-PHY Max Power */
	uint8_t			tssi_a;		/* Idle TSSI */
	uint8_t			tssi_bg;	/* Idle TSSI */
	uint16_t		bf_lo;		/* boardflags */
	uint16_t		bf_hi;		/* boardflags */
	struct {
		struct {
			int8_t a0, a1, a2, a3;
		} ghz24;
		struct {
			int8_t a0, a1, a2, a3;
		} ghz5;
	} again;	/* antenna gain */
};

struct ssb_info {
	struct ssb_board		boardinfo;
	struct ssb_sprom		sprom;
};

#define	SSB_DMA_TRANSLATION_MASK	0xc0000000

struct ssb_cc_pmu {
	uint8_t				rev;	/* PMU rev */
	uint32_t			freq;	/* crystal freq in kHz */
};

struct ssb_cc {
	struct ssb_dev_softc		*scc_dev;
	uint32_t			scc_caps;
	struct ssb_cc_pmu		scc_pmu;
	uint16_t			scc_powerup_delay;
};

struct ssb_pci {
	struct ssb_dev_softc		*spc_dev;
	uint8_t				spc_inited;
	uint8_t				spc_hostmode;
};

struct ssb_bus_ops {
	uint16_t			(*read16)(struct ssb_dev_softc *,
					    uint16_t);
	uint32_t			(*read32)(struct ssb_dev_softc *,
					    uint16_t);
	void				(*write16)(struct ssb_dev_softc *,
					    uint16_t, uint16_t);
	void				(*write32)(struct ssb_dev_softc *,
					    uint16_t, uint32_t);
	void				(*read_multi_1)(struct ssb_dev_softc *,
					    void *, size_t, uint16_t);
	void				(*read_multi_2)(struct ssb_dev_softc *,
					    void *, size_t, uint16_t);
	void				(*read_multi_4)(struct ssb_dev_softc *,
					    void *, size_t, uint16_t);
	void				(*write_multi_1)(struct ssb_dev_softc *,
					    const void *, size_t, uint16_t);
	void				(*write_multi_2)(struct ssb_dev_softc *,
					    const void *, size_t, uint16_t);
	void				(*write_multi_4)(struct ssb_dev_softc *,
					    const void *, size_t, uint16_t);
};

/* SSB device ID */
struct ssb_device_id {
	uint16_t			vendor;
	uint16_t			cid;
	uint8_t				rev;
};
#define	SSB_DEV(_vendor, _cid, _rev)	\
	{ SSB_VENDOR_##_vendor, SSB_DEV_##_cid, _rev, }

struct ssb_dev_softc {
	struct ssb_softc		*sd_bus;
	struct ssb_device_id		sd_id;
	const struct ssb_bus_ops	*sd_ops;

	uint8_t				sd_coreidx;
};

struct ssb_softc {
	device_t			ssb_dev;
	enum ssb_type			ssb_type;
	int				ssb_invalid;

	struct ssb_dev_softc		*ssb_curdev;	/* only for PCI */
	struct ssb_dev_softc		ssb_devs[SSB_MAX_CORES];
	int				ssb_ndevs;

	uint16_t			ssb_pci_vid;
	uint16_t			ssb_pci_did;
	uint16_t			ssb_pci_subvid;
	uint16_t			ssb_pci_subdid;
	int				ssb_mem_rid;
	struct resource			*ssb_mem_res;
	bus_space_tag_t			ssb_mem_bt;
	bus_space_handle_t		ssb_mem_bh;

	uint16_t			ssb_chipid;	/* for CORE 0 */
	uint16_t			ssb_chiprev;
	uint8_t				ssb_chippkg;

	struct ssb_cc			ssb_cc;		/* ChipCommon */
	struct ssb_pci			ssb_pci;	/* PCI-core */
	const struct ssb_bus_ops	*ssb_ops;

	struct ssb_board		ssb_board;	/* PCB */
	struct ssb_sprom		ssb_sprom;	/* SPROM */
	uint16_t			ssb_spromsize;	/* in word size */
};

int		ssb_attach(struct ssb_softc *);
int		ssb_detach(struct ssb_softc *);
void		ssb_powerup(struct ssb_softc *, int);
uint16_t	ssb_read_2(struct ssb_dev_softc *, uint16_t);
void		ssb_write_2(struct ssb_dev_softc *, uint16_t, uint16_t);
uint32_t	ssb_read_4(struct ssb_dev_softc *, uint16_t);
void		ssb_write_4(struct ssb_dev_softc *, uint16_t, uint32_t);
void		ssb_dev_up(struct ssb_dev_softc *, uint32_t);
void		ssb_dev_down(struct ssb_dev_softc *, uint32_t);
int		ssb_powerdown(struct ssb_softc *);
int		ssb_dev_isup(struct ssb_dev_softc *);
void		ssb_pcicore_intr(struct ssb_pci *, struct ssb_dev_softc *);
uint32_t	ssb_dma_translation(struct ssb_dev_softc *);
void		*ssb_dma_alloc_consistent(struct ssb_dev_softc *, size_t,
		    bus_addr_t *);
void		ssb_read_multi_1(struct ssb_dev_softc *, void *, size_t,
		    uint16_t);
void		ssb_read_multi_2(struct ssb_dev_softc *, void *, size_t,
		    uint16_t);
void		ssb_read_multi_4(struct ssb_dev_softc *, void *, size_t,
		    uint16_t);
void		ssb_write_multi_1(struct ssb_dev_softc *, const void *, size_t,
		    uint16_t);
void		ssb_write_multi_2(struct ssb_dev_softc *, const void *, size_t,
		    uint16_t);
void		ssb_write_multi_4(struct ssb_dev_softc *, const void *, size_t,
		    uint16_t);
void		ssb_barrier(struct ssb_dev_softc *, int);

#endif
